﻿using System;
using System.Drawing;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Threading;

namespace CM_2408C2E_CS
{
    public partial class Form1 : Form
    {
        /* Command ********************************************************** */
        private const byte CMD_ReadData = 0x81;               // 命令: 讀取資料
        private const byte CMD_L1_Set = 0x91;                 // 命令: L1設定值
        private const byte CMD_L2_Set = 0x92;                 // 命令: L2設定值
        private const byte CMD_L1_Output = 0xA1;              // 命令: L1輸出值
        private const byte CMD_L2_Output = 0xA2;              // 命令: L2輸出值
        private const byte CMD_L1_OnOff = 0xC1;               // 命令: L1 ON/OFF
        private const byte CMD_L2_OnOff = 0xC2;               // 命令: L2 ON/OFF
        private const byte CMD_L1_TrigOut = 0xD1;             // 命令: L1觸發輸出延遲
        private const byte CMD_L2_TrigOut = 0xD2;             // 命令: L2觸發輸出延遲
        private enum CMD { ReadData, L1_Set, L2_Set, L1_Output, L2_Output, L1_OnOff, L2_OnOff, L1_TrigOut, L2_TrigOut }
        /* Lx_Set =========================================================== */
        // bit7~4 最大電流
        private const byte SET_ExtOnOff = 0x02;                 // bit1   外部ON/OFF: 0=除能, 1=致能
        private const byte SET_OnOff = 0x01;                    // bit0   ON/OFF    : 0=OFF , 1=ON

        /* ACK ============================================================== */
        private const byte ACK_Failed = 0x80;                  // bit7 命令執行狀態: 0=成功, 1=失敗
        private const byte ACK_LowVolt = 0x10;                  // bit4 電源低電壓  : 0=No,   1=Yes
        private const byte ACK_L2_OC = 0x02;                  // bit1 L2過流      : 0=No,   1=Yes
        private const byte ACK_L1_OC = 0x01;                  // bit0 L1過流      : 0=No,   1=Yes

        /* ================================================================== */
        private const ushort LineMax = 256;                     // 最大行數

        /* ****************************************************************** */
        private static ushort LineNo = 0;                       // 行數編號
        private static ushort LineCount = 0;                    // 行數計數
        /* ------------------------------------------------------------------ */
        private static byte CNT_AskAuto = 0;                    // 計時-自動詢問
        private static byte CNT_Flash = 0;                      // 計時-閃爍
        /* 最大電流 --------------------------------------------------------- */
        private static int[] MaxCurrent = new int[] { 50, 100, 150, 200, 250, 300, 350, 400, 450, 500, 550, 600, 650, 700, 750, 800 };
        /* TX =============================================================== */
        private static byte[] Tx_Buffer;
        /* RX =============================================================== */
        private static byte[] RX_Buffer = new byte[8];         // 接收-封包
        private static byte RX_ACK = 0;                            // 回應碼
        private static byte ACK = 0;                            // 回應碼

        /* L1 =============================================================== */
        private static byte L1_Set = 0;                         // L1設定值
        private static int L1_Output = 0;                       // L1輸出值
        private static byte L1_TrigOut = 0;                     // L1觸發輸出延遲

        /* L2 =============================================================== */
        private static byte L2_Set = 0;                         // L2設定值
        private static int L2_Output = 0;                       // L2輸出值
        private static byte L2_TrigOut = 0;                     // L2觸發輸出延遲
        /* ****************************************************************** */
        private TcpClient client;
        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            cbb_L1_MaxC.SelectedIndex = 0;                      // L1最大電流 = 50mA
            lb_L1_MaxC.Text = Convert.ToString(MaxCurrent[0]);

            cbb_L2_MaxC.SelectedIndex = 0;                      // L2最大電流 = 50mA
            lb_L2_MaxC.Text = Convert.ToString(MaxCurrent[0]);
            ChangeTransmitStatus(false);
            var items = Properties.Settings.Default.HistoryIP;
            if(items!=null)
                foreach (var s in items) {
                    HistoryIPList.Items.Add(s);
                }
        }

        private void ChangeTransmitStatus(bool status)
        {
            if (status)
            {
                TransmitControlButton.ForeColor = Color.Red;
                TransmitControlButton.Text = "Disconnect";
                tss_Status.Text = "Connected";
                Lx_GroupBox_Enabled(true);                      // Lx群組方塊 (啟用)
            }
            else
            {
                TransmitControlButton.ForeColor = Color.Green;
                TransmitControlButton.Text = "Connect";
                tss_Status.Text = "Disconnected";
                Lx_GroupBox_Enabled(false);                      // Lx群組方塊 (不啟用)
            }
            TransmitControlButton.Enabled = true;
            ServerIP.Enabled = true;
            HistoryIPList.Enabled = true;
        }
        private void ConnectAction() {
            //連線成功:True, 失敗:False
            var status = ClientConnect(ServerIP.Text);
            MethodInvoker mi = new MethodInvoker(delegate {
                ChangeTransmitStatus(status);
            });
            this.BeginInvoke(mi, null);
        }
        private void TransmitControlButton_Click(object sender, EventArgs e)
        {
            ServerIP.Text = ServerIP.Text.Trim();
            ServerIP.Enabled = false;
            TransmitControlButton.Enabled = false;
            HistoryIPList.Enabled = false;
            //正處於連線狀態 -> 關閉連線
            if (client != null)
            {
                client.Close();
                client = null;
                Dsp_String(Color.AliceBlue, "IP: " + ServerIP.Text + " 已中斷連線!");
                ChangeTransmitStatus(false);
            }
            //未連線 -> 對輸入的IP發起連線請求
            else
            {
                Thread thread = new Thread(new ThreadStart(ConnectAction));
                thread.Start();
            }
            if (!HistoryIPList.Items.Contains(ServerIP.Text)) HistoryIPList.Items.Add(ServerIP.Text);
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            timer.Interval = 50;
            if (client != null && client.Connected) {
                Ask_Auto();// 輪詢更新狀態  
            }
        }

        private void UpdateLxUI()
        {
            Update_L1(new byte[] { RX_Buffer[0], RX_Buffer[1], RX_Buffer[2], RX_Buffer[3] });   // 更新-L1參數
            Update_L2(new byte[] { RX_Buffer[4], RX_Buffer[5], RX_Buffer[6], RX_Buffer[7] });   // 更新-L2參數
            Check_ACK();                            // 檢查-回應碼
        }

        /* ================================================================== */

        private void Ask_Auto()                                 // 自動詢問
        {
            if (cbb_L1_MaxC.DroppedDown == true) { CNT_AskAuto = 0; return; }
            if (cbb_L2_MaxC.DroppedDown == true) { CNT_AskAuto = 0; return; }

            if (nud_L1_Output.Focused == true) { CNT_AskAuto = 0; return; }
            if (nud_L2_Output.Focused == true) { CNT_AskAuto = 0; return; }

            if (nud_L1_TrigOut.Focused == true) { CNT_AskAuto = 0; return; }
            if (nud_L2_TrigOut.Focused == true) { CNT_AskAuto = 0; return; }

            if (++CNT_AskAuto >= 10)                            // 自動刷新 (10 * 50ms = 0.5sec)
            {
                SendCommend(CMD.ReadData);
                CNT_AskAuto = 0;
                if ((ACK & 0x13) != 0)                              // 電源低電壓/L2過流/L1過流 ?
                {
                    Lx_GroupBox_Enabled(false);                     // Lx群組方塊 (不啟用)
                    Alarm_Flash();                                  // 警告訊息-閃爍 (2Hz)
                }
                else
                {
                    Lx_GroupBox_Enabled(true);                      // Lx群組方塊 (啟用)
                }
            }
        }

        private void Update_L1(byte[] L1_Packet)                                // 更新-L1參數
        {
            L1_Set = (byte)(L1_Packet[0] & 0xF3);

            if ((L1_Set & SET_OnOff) == 0)                      // L1 OFF ?
            {
                lb_L1_LED.BackColor = Color.DimGray;
                btn_L1_OnOff.Text = "ON";
            }
            else                                                // L1 ON
            {
                lb_L1_LED.BackColor = Color.Green;
                btn_L1_OnOff.Text = "OFF";
            }

            if ((L1_Set & SET_ExtOnOff) == 0)                   // L1外部ON/OFF-除能?
            {
                ckb_L1_ExtOnOff.Checked = false;
                lb_L1_LED.Visible = true;
                btn_L1_OnOff.Visible = true;
            }
            else                                                // L1外部ON/OFF-致能
            {
                ckb_L1_ExtOnOff.Checked = true;
                lb_L1_LED.Visible = false;
                btn_L1_OnOff.Visible = false;
            }

            /* L1最大電流 --------------------------------------------------- */
            cbb_L1_MaxC.SelectedIndex = (L1_Set >> 4) & 0x0F;
            int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];

            lb_L1_MaxC.Text = Convert.ToString(MaxC);
            tkb_L1_Output.TickFrequency = MaxC / 10;
            nud_L1_Output.Maximum = MaxC;
            tkb_L1_Output.Maximum = MaxC;

            /* L1輸出值 ----------------------------------------------------- */
            int Output = L1_Packet[1];
            Output <<= 8;
            Output |= L1_Packet[2];

            nud_L1_Output.Value = Output;
            tkb_L1_Output.Value = Output;
            L1_Output = Output;
            lb_L1_PCV.Text = String.Format("{0:F1}%", (float)(Output * 100) / MaxC);
            /* L1觸發輸出延遲 ----------------------------------------------- */
            L1_TrigOut = L1_Packet[3];
            nud_L1_TrigOut.Value = L1_TrigOut;
            lb_L1_TrigOut.Text = Convert.ToString(L1_TrigOut * 5) + "ms";
        }

        private void Update_L2(byte[] L2_Packet)                                // 更新-L2參數
        {
            L2_Set = (byte)(L2_Packet[0] & 0xF3);

            if ((L2_Set & SET_OnOff) == 0)                      // L2 OFF ?
            {
                lb_L2_LED.BackColor = Color.DimGray;
                btn_L2_OnOff.Text = "ON";
            }
            else                                                // L2 ON
            {
                lb_L2_LED.BackColor = Color.Green;
                btn_L2_OnOff.Text = "OFF";
            }

            if ((L2_Set & SET_ExtOnOff) == 0)                   // L2外部ON/OFF-除能?
            {
                ckb_L2_ExtOnOff.Checked = false;
                lb_L2_LED.Visible = true;
                btn_L2_OnOff.Visible = true;
            }
            else                                                // L2外部ON/OFF-致能
            {
                ckb_L2_ExtOnOff.Checked = true;
                lb_L2_LED.Visible = false;
                btn_L2_OnOff.Visible = false;
            }

            /* L2最大電流 --------------------------------------------------- */
            cbb_L2_MaxC.SelectedIndex = (L2_Set >> 4) & 0x0F;
            int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];

            lb_L2_MaxC.Text = Convert.ToString(MaxC);
            tkb_L2_Output.TickFrequency = MaxC / 10;
            nud_L2_Output.Maximum = MaxC;
            tkb_L2_Output.Maximum = MaxC;

            /* L2輸出值 ----------------------------------------------------- */
            int Output = L2_Packet[1];
            Output <<= 8;
            Output |= L2_Packet[2];

            nud_L2_Output.Value = Output;
            tkb_L2_Output.Value = Output;
            L2_Output = Output;
            lb_L2_PCV.Text = String.Format("{0:F1}%", (float)(Output * 100) / MaxC);

            /* L2觸發輸出延遲 ----------------------------------------------- */
            L2_TrigOut = L2_Packet[3];
            nud_L2_TrigOut.Value = L2_TrigOut;
            lb_L2_TrigOut.Text = Convert.ToString(L2_TrigOut * 5) + "ms";
        }

        private void Check_ACK()                                // 檢查-回應碼
        {
            ACK = (byte)(RX_ACK & 0x93);
            if ((ACK & ACK_Failed) == ACK_Failed)               // bit7: 命令執行失敗?
            {
                Dsp_String(Color.Red, " 命令執行失敗!");
            }

            if ((ACK & ACK_LowVolt) == ACK_LowVolt)             // bit4: 電源低電壓?
            {
                Dsp_String(Color.Red, " 電源低電壓!");
            }
            else lb_LowVolt.BackColor = Color.DimGray;

            if ((ACK & 0x03) == 0x03)                           // L2過流 & L1過流 ?
            {
                Dsp_String(Color.Red, " L1-L2過流!");
            }
            else
            {
                if ((ACK & ACK_L2_OC) == ACK_L2_OC)             // bit1: L2過流?
                {
                    Dsp_String(Color.Red, " L2過流!");
                }
                else lb_L2_OC.BackColor = Color.DimGray;

                if ((ACK & ACK_L1_OC) == ACK_L1_OC)             // bit0: L1過流?
                {
                    Dsp_String(Color.Red, " L1過流!");
                }
                else lb_L1_OC.BackColor = Color.DimGray;
            }
        }

        private void Alarm_Flash()                              // 警告訊息-閃爍 (2Hz)
        {
            if ((ACK & ACK_LowVolt) == ACK_LowVolt)             // bit4: 電源低電壓?
            {
                if (CNT_Flash == 0) lb_LowVolt.BackColor = Color.Red;
                else lb_LowVolt.BackColor = Color.DimGray;
            }

            if ((ACK & ACK_L2_OC) == ACK_L2_OC)                 // bit1: L2過流?
            {
                if (CNT_Flash == 0) lb_L2_OC.BackColor = Color.Red;
                else lb_L2_OC.BackColor = Color.DimGray;
            }

            if ((ACK & ACK_L1_OC) == ACK_L1_OC)                 // bit0: L1過流?
            {
                if (CNT_Flash == 0) lb_L1_OC.BackColor = Color.Red;
                else lb_L1_OC.BackColor = Color.DimGray;
            }

            if (CNT_Flash++ > 0) CNT_Flash = 0;                 // 10 * 50ms = 500ms
        }

        /* RichTextBox 顯示 ************************************************* */

        private void Dsp_String(Color text_color, string str)   // 顯示字串
        {
            if (++LineCount > LineMax)                          // 行數計數 > 最大行數?
            {
                LineCount = 1;
                richTextBox.Clear();
            }

            if (LineCount > 1) richTextBox.AppendText("\n");
            richTextBox.SelectionStart = richTextBox.Text.Length;
            richTextBox.ScrollToCaret();

            /* 行數編號顯示 ------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.DimGray;
            richTextBox.SelectionColor = Color.White;
            richTextBox.AppendText(String.Format("{0:X4}", LineNo++));

            /* 字串顯示 ----------------------------------------------------- */
            richTextBox.SelectionBackColor = Color.Black;
            richTextBox.SelectionColor = text_color;
            richTextBox.AppendText(str);
            richTextBox.Select();
        }

        private void Lx_GroupBox_Enabled(bool s)                // Lx群組方塊啟用否?
        {
            gpb_L1_OnOffCtrl.Enabled = s;
            gpb_L1_Output.Enabled = s;
            gpb_L1_TrigOut.Enabled = s;

            gpb_L2_OnOffCtrl.Enabled = s;
            gpb_L2_Output.Enabled = s;
            gpb_L2_TrigOut.Enabled = s;
        }

        /* L1輸出調整 ******************************************************* */

        private void cbb_L1_MaxC_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (((L1_Set >> 4) & 0x0F) != cbb_L1_MaxC.SelectedIndex)
            {
                byte i = (byte)(L1_Set & (SET_ExtOnOff | SET_OnOff));
                L1_Set = (byte)((cbb_L1_MaxC.SelectedIndex << 4) | i);

                nud_L1_Output.Value = 0;
                tkb_L1_Output.Value = 0;
                L1_Output = 0;

                int MaxC = MaxCurrent[cbb_L1_MaxC.SelectedIndex];
                lb_L1_MaxC.Text = Convert.ToString(MaxC);
                tkb_L1_Output.TickFrequency = MaxC / 10;
                nud_L1_Output.Maximum = MaxC;
                tkb_L1_Output.Maximum = MaxC;

                SendCommend(CMD.L1_Set);
            }
        }

        private void tkb_L1_Output_ValueChanged(object sender, EventArgs e)
        {
            nud_L1_Output.Value = tkb_L1_Output.Value;
            L1_Output = tkb_L1_Output.Value;
            lb_L1_PCV.Text = String.Format("{0:F1}%", (float)(L1_Output * 100) / MaxCurrent[cbb_L1_MaxC.SelectedIndex]);

            SendCommend(CMD.L1_Output);
        }

        private void nud_L1_Output_ValueChanged(object sender, EventArgs e)
        {
            tkb_L1_Output.Value = (int)nud_L1_Output.Value;
            L1_Output = tkb_L1_Output.Value;
            lb_L1_PCV.Text = String.Format("{0:F1}%", (float)(L1_Output * 100) / MaxCurrent[cbb_L1_MaxC.SelectedIndex]);

            SendCommend(CMD.L1_Output);
        }

        /* L1 ON/OFF ======================================================== */

        private void ckb_L1_ExtOnOff_CheckedChanged(object sender, EventArgs e)
        {
            if (ckb_L1_ExtOnOff.Checked == false)               // Enable -> Disable ?
            {
                L1_Set &= (SET_ExtOnOff ^ 0xFF);                // L1 EXT. ON/OFF = Disable
                lb_L1_LED.Visible = true;
                btn_L1_OnOff.Visible = true;
            }
            else                                                // Disable -> Enable
            {
                L1_Set |= SET_ExtOnOff;                         // L1 EXT. ON/OFF = Enable
                lb_L1_LED.Visible = false;
                btn_L1_OnOff.Visible = false;
            }

            SendCommend(CMD.L1_Set);
        }

        private void btn_L1_OnOff_Click(object sender, EventArgs e)
        {
            if ((L1_Set & SET_OnOff) == 0)                      // OFF -> ON ?
            {
                L1_Set |= SET_OnOff;                            // L1 ON
                lb_L1_LED.BackColor = Color.Green;
                btn_L1_OnOff.Text = "OFF";
            }
            else                                                // ON -> OFF
            {
                L1_Set &= (SET_OnOff ^ 0xFF);                   // L1 OFF
                lb_L1_LED.BackColor = Color.DimGray;
                btn_L1_OnOff.Text = "ON";
            }

            SendCommend(CMD.L1_OnOff);

        }

        /* L1觸發輸出延遲 =================================================== */

        private void nud_L1_TrigOut_ValueChanged(object sender, EventArgs e)
        {
            L1_TrigOut = (byte)nud_L1_TrigOut.Value;
            lb_L1_TrigOut.Text = Convert.ToString(L1_TrigOut * 5) + "ms";

            SendCommend(CMD.L1_TrigOut);
        }

        /* L2輸出調整 ******************************************************* */

        private void cbb_L2_MaxC_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (((L2_Set >> 4) & 0x0F) != cbb_L2_MaxC.SelectedIndex)
            {
                byte i = (byte)(L2_Set & (SET_ExtOnOff | SET_OnOff));
                L2_Set = (byte)((cbb_L2_MaxC.SelectedIndex << 4) | i);

                nud_L2_Output.Value = 0;
                tkb_L2_Output.Value = 0;
                L2_Output = 0;

                int MaxC = MaxCurrent[cbb_L2_MaxC.SelectedIndex];
                lb_L2_MaxC.Text = Convert.ToString(MaxC);
                tkb_L2_Output.TickFrequency = MaxC / 10;
                nud_L2_Output.Maximum = MaxC;
                tkb_L2_Output.Maximum = MaxC;

                SendCommend(CMD.L2_Set);
            }
        }

        private void tkb_L2_Output_ValueChanged(object sender, EventArgs e)
        {
            nud_L2_Output.Value = tkb_L2_Output.Value;
            L2_Output = tkb_L2_Output.Value;
            lb_L2_PCV.Text = String.Format("{0:F1}%", (float)(L2_Output * 100) / MaxCurrent[cbb_L2_MaxC.SelectedIndex]);

            SendCommend(CMD.L2_Output);
        }

        private void nud_L2_Output_ValueChanged(object sender, EventArgs e)
        {
            tkb_L2_Output.Value = (int)nud_L2_Output.Value;
            L2_Output = tkb_L2_Output.Value;
            lb_L2_PCV.Text = String.Format("{0:F1}%", (float)(L2_Output * 100) / MaxCurrent[cbb_L2_MaxC.SelectedIndex]);

            SendCommend(CMD.L2_Output);
        }

        /* L2 ON/OFF ======================================================== */

        private void ckb_L2_ExtOnOff_CheckedChanged(object sender, EventArgs e)
        {
            if (ckb_L2_ExtOnOff.Checked == false)               // Enable -> Disable ?
            {
                L2_Set &= (SET_ExtOnOff ^ 0xFF);                // L2 EXT. ON/OFF = Disable
                lb_L2_LED.Visible = true;
                btn_L2_OnOff.Visible = true;
            }
            else                                                // Disable -> Enable
            {
                L2_Set |= SET_ExtOnOff;                         // L2 EXT. ON/OFF = Enable
                lb_L2_LED.Visible = false;
                btn_L2_OnOff.Visible = false;
            }

            SendCommend(CMD.L2_Set);
        }

        private void btn_L2_OnOff_Click(object sender, EventArgs e)
        {
            if ((L2_Set & SET_OnOff) == 0)                      // OFF -> ON
            {
                L2_Set |= SET_OnOff;                            // L2 ON
                lb_L2_LED.BackColor = Color.Green;
                btn_L2_OnOff.Text = "OFF";
            }
            else                                                // ON -> OFF
            {
                L2_Set &= (SET_OnOff ^ 0xFF);                   // L2 OFF
                lb_L2_LED.BackColor = Color.DimGray;
                btn_L2_OnOff.Text = "ON";
            }

            SendCommend(CMD.L2_OnOff);
        }

        /* L2觸發輸出延遲 =================================================== */

        private void nud_L2_TrigOut_ValueChanged(object sender, EventArgs e)
        {
            L2_TrigOut = (byte)nud_L2_TrigOut.Value;
            lb_L2_TrigOut.Text = Convert.ToString(L2_TrigOut * 5) + "ms";

            SendCommend(CMD.L2_TrigOut);
        }
        /* 網路連線 *********************************************************** */
        private bool ClientConnect(string ServerIP)
        {
            try
            {
                //建立TCP/IP協定連線 Client模式
                client = new TcpClient();
                MethodInvoker mi = new MethodInvoker(delegate {
                    tss_Status.Text = "Connecting...";
                });
                this.BeginInvoke(mi, null);
                

                //5秒超時
                client.ConnectAsync(ServerIP, 5000).Wait(5000);


                if (client.Connected)
                {
                    mi = new MethodInvoker(delegate {
                        Dsp_String(Color.AliceBlue, "IP: " + ServerIP + " 連線成功!");
                    });
                    this.BeginInvoke(mi, null);
                    
                    return true;
                }
                else
                {
                    client = null;
                    mi = new MethodInvoker(delegate {
                        Dsp_String(Color.Red, "連線超時!");
                    });
                    this.BeginInvoke(mi, null);
                   
                    return false;
                }

            }
            catch (Exception ex)
            {
                client = null;
                MethodInvoker mi = new MethodInvoker(delegate {
                    Dsp_String(Color.Red, "發生未預期錯誤!");
                });
                this.BeginInvoke(mi, null);
                Console.WriteLine(ex.StackTrace);
                return false;
            }
        }
        //封包指令-字串轉換(顯示用)
        private string getCommendString(byte commend)
        {
            switch (commend)
            {
                case CMD_ReadData: return "ReadData";
                case CMD_L1_Set: return "L1_Set";
                case CMD_L2_Set: return "L2_Set";
                case CMD_L1_Output: return "L1_Output";
                case CMD_L2_Output: return "L2_Output";
                case CMD_L1_OnOff: return "L1_OnOff";
                case CMD_L2_OnOff: return "L2_OnOff";
                case CMD_L1_TrigOut: return "L1_TrigOut";
                case CMD_L2_TrigOut: return "L2_TrigOut";
                default: return "Error";
            }
        }
        //顯示傳送封包資料
        private void Dsp_BufferData(byte[] buffer, bool isTX)
        {
            string str = isTX ? "TX:" : "RX:";
            string commend = getCommendString(buffer[0]);
            //顯示16進制封包
            for (int i = 0; i < buffer.Length; i++) str += string.Format(" {0:X2}", buffer[i]);
            Dsp_String(Color.AliceBlue, commend + ", " + str);
        }
        private void SendToServer(byte[] data, int response_size) {
            if (client.Connected)
            {
                byte[] read = new byte[response_size];
                NetworkStream serverStream = client.GetStream();
                try {
                    // RX超時設定500ms
                    serverStream.ReadTimeout = 500;
                    serverStream.Write(data, 0, data.Length);
                    serverStream.Read(read, 0, read.Length);
                    // 封包處理
                    OnDataReceived(read, read.Length);
                }

                catch (ObjectDisposedException ex) {
                    Dsp_String(Color.Red, "發生未預期錯誤!");
                }
                catch (ArgumentNullException ex) {
                    Dsp_String(Color.Red, "發生未預期錯誤!");
                }
                catch (ArgumentOutOfRangeException ex)
                {
                    Dsp_String(Color.Red, "發生未預期錯誤!");
                }
                catch (System.IO.IOException ex)
                {
                    Dsp_String(Color.Red, "連線沒有回應!");
                }
            }
        }
        private void SendCommend(CMD commend)
        {
            int response_size = 3;
            byte Checksum;
            switch (commend)
            {
                case CMD.ReadData:
                    Checksum = CMD_ReadData;
                    Tx_Buffer = new byte[] { CMD_ReadData, Checksum};
                    response_size = 11;
                    break;
                case CMD.L1_Set:
                    Checksum = (byte)(CMD_L1_Set + L1_Set);
                    Tx_Buffer = new byte[] { CMD_L1_Set, L1_Set, Checksum };
                    break;
                case CMD.L2_Set:
                    Checksum = (byte)(CMD_L2_Set + L2_Set);
                    Tx_Buffer = new byte[] { CMD_L2_Set, L2_Set, Checksum };
                    break;
                case CMD.L1_Output:
                    Checksum = (byte)(CMD_L1_Output + (byte)(L1_Output >> 8) + (byte)L1_Output);
                    Tx_Buffer = new byte[] { CMD_L1_Output, (byte)(L1_Output >> 8), (byte)L1_Output, Checksum };
                    break;
                case CMD.L2_Output:
                    Checksum = (byte)(CMD_L2_Output + (byte)(L2_Output >> 8) + (byte)L2_Output);
                    Tx_Buffer = new byte[] { CMD_L2_Output, (byte)(L2_Output >> 8), (byte)L2_Output, Checksum };
                    break;
                case CMD.L1_OnOff:
                    Checksum = (byte)(CMD_L1_OnOff + (byte)(L1_Set & SET_OnOff));
                    Tx_Buffer = new byte[] { CMD_L1_OnOff, (byte)(L1_Set & SET_OnOff), Checksum };// L1 ON/OFF
                    break;
                case CMD.L2_OnOff:
                    Checksum = (byte)(CMD_L2_OnOff + (byte)(L2_Set & SET_OnOff));
                    Tx_Buffer = new byte[] { CMD_L2_OnOff, (byte)(L2_Set & SET_OnOff), Checksum };// L2 ON/OFF
                    break;
                case CMD.L1_TrigOut:
                    Checksum = (byte)(CMD_L1_TrigOut + L1_TrigOut);
                    Tx_Buffer = new byte[] { CMD_L1_TrigOut, L1_TrigOut, Checksum };// L1 Trig
                    break;
                case CMD.L2_TrigOut:
                    Checksum = (byte)(CMD_L2_TrigOut + L2_TrigOut);
                    Tx_Buffer = new byte[] { CMD_L2_TrigOut, L2_TrigOut, Checksum };// L1 Trig
                    break;
            }
            SendToServer(Tx_Buffer, response_size);
            Dsp_BufferData(Tx_Buffer, true);
        }
        // 讀到資料後，更新UI
        public void OnDataReceived(byte[] data, int length)
        {
            byte CMD = data[0];
            if (CMD == ACK_Failed) length = 3;
            byte Checksum = 0;
            //Console.WriteLine("Data:" + BitConverter.ToString(data));
            for (int i = 0; i < length - 1; i++)
            {
                Checksum += data[i];
            }
            //Console.WriteLine("Checksum:" + BitConverter.ToString(new byte[] { Checksum }));
            if (Checksum == data[length - 1])
            {
                MethodInvoker mi = new MethodInvoker(delegate {
                    Dsp_BufferData(data, false);
                });
                this.BeginInvoke(mi, null);
                if (data[0] == ACK_Failed)
                {
                    RX_ACK = data[1];
                    //Console.WriteLine("RX_ACK:" + RX_ACK);
                }
                else if (data[0] == CMD_ReadData)
                {
                    RX_Buffer = new byte[] { data[1], data[2], data[3], data[4], data[5], data[6], data[7], data[8] };
                    RX_ACK = data[9];
                    mi = new MethodInvoker(this.UpdateLxUI);
                    this.BeginInvoke(mi, null);
                    //Console.WriteLine("Receive Success");
                }
                else
                {
                    RX_ACK = data[length - 2];
                    //Console.WriteLine("RX_ACK:" + RX_ACK);
                }
                mi = new MethodInvoker(Check_ACK);
                this.BeginInvoke(mi, null);
            }
            else
            {
                MethodInvoker mi = new MethodInvoker(delegate {
                    Dsp_BufferData(data, false);
                });
                this.BeginInvoke(mi, null);
                //Console.WriteLine("Receive Fail");
            }

        }

        private void HistoryIPList_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (HistoryIPList.SelectedIndex != -1) {
                ServerIP.Text = (string) HistoryIPList.SelectedItem;
            }
        }

        private void Form1_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (HistoryIPList.Items.Count > 0) {
                Properties.Settings.Default.HistoryIP = new System.Collections.Specialized.StringCollection();
                foreach (var s in HistoryIPList.Items) {
                    Properties.Settings.Default.HistoryIP.Add((string)s);
                    Properties.Settings.Default.Save();
                }
            }
        }

        private void ServerIP_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                TransmitControlButton.PerformClick();
            }
        }

        private void HistoryIPList_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            int index = this.HistoryIPList.IndexFromPoint(e.Location);
            if (index != -1) {
                DialogResult result = MessageBox.Show($"是否要刪除IP：{HistoryIPList.Items[index]}", "提醒", MessageBoxButtons.YesNo);
                if (result == DialogResult.Yes) HistoryIPList.Items.RemoveAt(index);
            }
        }
    }
}
